# from pathlib import Path
# from unittest.mock import Mock

# from azure_helper.steps.deploy_aml_model import DeploymentSettings, DeployModel

# test_module = "azure_helper.steps.deploy_aml_model"


# class TestDeployAMLModel:
#     def test_get_inference_config(
#         self,
#         mocker,
#     ):

#         # We are not testing Azure ML SDK functionality in a unit test
#         # Rather that the correct calls are made

#         mock_inference_config = mocker.patch(f"{test_module}.InferenceConfig")
#         mock_inference_config_obj = Mock()
#         mock_config = Mock()
#         mock_inference_config_obj.return_value = mock_config

#         mock_env = mocker.patch(f"{test_module}.Environment")

#         mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")

#         workspace_name = "test_workspace"
#         resource_group = "test_rg"
#         subscription_id = "test_sub_id"

#         spn_credentials = {
#             "tenant_id": "test_tenant_id",
#             "service_principal_id": "test_spn_id",
#             "service_principal_password": "test_spn_passwd",
#         }

#         aml_interface = mock_aml_inteface(
#             spn_credentials,
#             subscription_id,
#             workspace_name,
#             resource_group,
#         )

#         settings = DeploymentSettings(
#             deployment_service_name="test_deployment_service_name",
#             cpu_cores=1,
#             gpu_cores=0,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         deployment = DeployModel(
#             aml_interface=aml_interface,
#             aml_env_name="test_aml_env_name",
#             model_name="test_model_name",
#             script_config_path=Path("test_data/test_score.py"),
#             deployment_settings=settings,
#         )

#         mock_config = deployment.get_inference_config()

#         mock_env.get.assert_called_once()
#         mock_inference_config.assert_called_once()

#     def test_deploy_aciservice(
#         self,
#         mocker,
#     ):
#         mock_get_inference_config = mocker.patch(
#             f"{test_module}.DeployModel.get_inference_config",
#         )
#         mock_config = Mock()
#         mock_get_inference_config.return_value = mock_config

#         mock_AciWebservice = mocker.patch(f"{test_module}.AciWebservice")
#         mock_Model = mocker.patch(f"{test_module}.Model")
#         mock_service = Mock()
#         mock_service.scoring_uri = "https://foo.bar/"
#         mock_Model.deploy.return_value = mock_service

#         mock_env = mocker.patch(f"{test_module}.Environment")

#         mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")

#         workspace_name = "test_workspace"
#         resource_group = "test_rg"
#         subscription_id = "test_sub_id"

#         spn_credentials = {
#             "tenant_id": "test_tenant_id",
#             "service_principal_id": "test_spn_id",
#             "service_principal_password": "test_spn_passwd",
#         }

#         aml_interface = mock_aml_inteface(
#             spn_credentials,
#             subscription_id,
#             workspace_name,
#             resource_group,
#         )

#         settings = DeploymentSettings(
#             deployment_service_name="test_deployment_service_name",
#             cpu_cores=1,
#             gpu_cores=0,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         deployment = DeployModel(
#             aml_interface=aml_interface,
#             aml_env_name="test_aml_env_name",
#             model_name="test_model_name",
#             script_config_path=Path("test_data/test_score.py"),
#             deployment_settings=settings,
#         )

#         deployment.deploy_aciservice()

#         mock_get_inference_config.assert_called_once()
#         mock_AciWebservice.deploy_configuration.assert_called_once_with(
#             cpu_cores=1,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         mock_Model.deploy.assert_called_once()

#         mock_service.wait_for_deployment.assert_called_once()

#     def test_deploy_aksservice(self, mocker, caplog):

#         aks_cluster_name = "test_aks_cluster_name"

#         mock_ComputeTarget = mocker.patch(f"{test_module}.ComputeTarget")

#         mock_get_inference_config = mocker.patch(
#             f"{test_module}.DeployModel.get_inference_config",
#         )
#         mock_config = Mock()
#         mock_get_inference_config.return_value = mock_config

#         mock_AksWebservice = mocker.patch(f"{test_module}.AksWebservice")

#         mock_Model = mocker.patch(f"{test_module}.Model")
#         mock_service = Mock()
#         mock_service.scoring_uri = "https://foo.bar/"
#         mock_Model.deploy.return_value = mock_service

#         mock_env = mocker.patch(f"{test_module}.Environment")

#         mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")

#         workspace_name = "test_workspace"
#         resource_group = "test_rg"
#         subscription_id = "test_sub_id"

#         spn_credentials = {
#             "tenant_id": "test_tenant_id",
#             "service_principal_id": "test_spn_id",
#             "service_principal_password": "test_spn_passwd",
#         }

#         aml_interface = mock_aml_inteface(
#             spn_credentials,
#             subscription_id,
#             workspace_name,
#             resource_group,
#         )

#         settings = DeploymentSettings(
#             deployment_service_name="test_deployment_service_name",
#             cpu_cores=1,
#             gpu_cores=0,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         deployment = DeployModel(
#             aml_interface=aml_interface,
#             aml_env_name="test_aml_env_name",
#             model_name="test_model_name",
#             script_config_path=Path("test_data/test_score.py"),
#             deployment_settings=settings,
#         )

#         deployment.deploy_aksservice(aks_cluster_name=aks_cluster_name)

#         mock_get_inference_config.assert_called_once()
#         mock_AksWebservice.deploy_configuration.assert_called_once_with(
#             cpu_cores=1,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         mock_Model.deploy.assert_called_once()

#         mock_service.wait_for_deployment.assert_called_once()

#     # def test_deploy_aksservice_no_cluster(
#     #     self,
#     #     mocker,
#     #     caplog,
#     # ):
#     #     aks_cluster_name = "test_aks_cluster_name"

#     #     mock_compute_target = mocker.patch(f"{test_module}.ComputeTarget")

#     #     mock_compute = Mock()
#     #     mock_compute_target.create.return_value = mock_compute

#     #     mock_aks_compute = mocker.patch(f"{test_module}.AksCompute")

#     #     mock_get_inference_config = mocker.patch(
#     #         f"{test_module}.DeployModel.get_inference_config",
#     #     )
#     #     mock_config = Mock()
#     #     mock_get_inference_config.return_value = mock_config

#     #     mock_aks_webservice = mocker.patch(f"{test_module}.AksWebservice")

#     #     mock_Model = mocker.patch(f"{test_module}.Model")
#     #     mock_service = Mock()
#     #     mock_service.scoring_uri = "https://foo.bar/"
#     #     mock_Model.deploy.return_value = mock_service

#     #     mock_env = mocker.patch(f"{test_module}.Environment")

#     #     mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")

#     #     workspace_name = "test_workspace"
#     #     resource_group = "test_rg"
#     #     subscription_id = "test_sub_id"

#     #     spn_credentials = {
#     #         "tenant_id": "test_tenant_id",
#     #         "service_principal_id": "test_spn_id",
#     #         "service_principal_password": "test_spn_passwd",
#     #     }

#     #     aml_interface = mock_aml_inteface(
#     #         spn_credentials,
#     #         subscription_id,
#     #         workspace_name,
#     #         resource_group,
#     #     )

#     #     settings = DeploymentSettings(
#     #         deployment_service_name="test_deployment_service_name",
#     #         cpu_cores=1,
#     #         gpu_cores=0,
#     #         memory_gb=1,
#     #         enable_app_insights=True,
#     #     )

#     #     deployment = DeployModel(
#     #         aml_interface=aml_interface,
#     #         aml_env_name="test_aml_env_name",
#     #         model_name="test_model_name",
#     #         script_config_path=Path("test_data/test_score.py"),
#     #         deployment_settings=settings,
#     #     )

#     #     mock_compute_target.side_effect = ComputeTargetException

#     #     deployment.deploy_aksservice(
#     #         aks_cluster_name=aks_cluster_name,
#     #     )

#     #     assert (
#     #         f"k8s cluster {aks_cluster_name} was not found in workspace test_workspace. Now provisioning one."
#     #         in caplog.records[0].message
#     #     )

#     def test_update_service(self, mocker):

#         mock_get_inference_config = mocker.patch(
#             f"{test_module}.DeployModel.get_inference_config",
#         )
#         mock_config = Mock()
#         mock_get_inference_config.return_value = mock_config

#         mock_aml_interface = mocker.patch(f"{test_module}.AMLInterface")

#         mock_Webservice = mocker.patch(f"{test_module}.Webservice")
#         mock_Webservice_obj = Mock()
#         mock_Webservice.return_value = mock_Webservice_obj

#         workspace_name = "test_workspace"
#         resource_group = "test_rg"
#         subscription_id = "test_sub_id"

#         spn_credentials = {
#             "tenant_id": "test_tenant_id",
#             "service_principal_id": "test_spn_id",
#             "service_principal_password": "test_spn_passwd",
#         }

#         aml_interface = mock_aml_interface(
#             spn_credentials,
#             subscription_id,
#             workspace_name,
#             resource_group,
#         )
#         aml_interface.workspace.models.get.return_value = "test_model_name"

#         settings = DeploymentSettings(
#             deployment_service_name="test_deployment_service_name",
#             cpu_cores=1,
#             gpu_cores=0,
#             memory_gb=1,
#             enable_app_insights=True,
#         )

#         deployment = DeployModel(
#             aml_interface=aml_interface,
#             aml_env_name="test_aml_env_name",
#             model_name="test_model_name",
#             script_config_path=Path("test_data/test_score.py"),
#             deployment_settings=settings,
#         )

#         deployment.update_service()
#         mock_get_inference_config.assert_called_once()
#         mock_Webservice_obj.update.assert_called_once_with(
#             models=["test_model_name"],
#             inference_config=mock_config,
#         )


# # @patch("src.my_custom_package.deploy_aml_model.Webservice")
# # @patch("src.my_custom_package.deploy_aml_model.get_inference_config")
# # def test_update_service(mock_get_inference_config, mock_Webservice):
# #     mock_aml_interface = Mock()
# #     mock_get_inference_config.return_value = "test_inference_config"
# #     mock_aml_interface.workspace.models.get.return_value = "test_model_name"
# #     mock_service = Mock()
# #     mock_Webservice.return_value = mock_service

# #     update_service(mock_aml_interface)
# #     # mock_get_inference_config is tested independently above
# #     mock_get_inference_config.assert_called_once()
# #     mock_service.update.assert_called_once_with(
# #         models=["test_model_name"],
# #         inference_config="test_inference_config",
# #     )



from azure.ai.ml.entities import BuildContext, Environment

from azure_helper.logger import get_logger

log = get_logger()


class AMLEnvironment:
    def __init__(self, image: str, dockerfile_path: str = "Dockerfile") -> None:
        """Instantiate the creation of the AzureML Environment needed for the experiment.

        An AzureML Environment is a Docker Image which encapsulates all the needed requirements for the experiment (ie
        the training of a model) to run, this can be :

        * the version of the OS,
        * the various python libraries needed (ie the `requirements.txt`),
        * the project built as a wheel and added as a private pip package,
        * third-party softwares, etc.

        To create the Docker image of the environment, we use a base image managed by Microsoft. This image can be changed corresponding to the
        need of the experiment.

        The standard image we use is `mcr.microsoft.com/azureml/curated/sklearn-1.0-ubuntu20.04-py38-cpu`.

        The different images managed by Microsoft can be found ine the [Azure Machine Learning base images](https://github.com/Azure/AzureML-Containers).

        The create the Docker image of the environment, we also need the project, built as a pip wheel (by Flit, setuptools, Hatch, etc),
        so we can install it in our environment.

        Args:
            dist_dir (Path): The path where the project built as a wheel is located.
            base_image (str): The base Docker image we use to create the environment.
        """
        # self.private_wheel_location = private_wheel_location
        self.base_image = image
        self.dockerfile_path = dockerfile_path

    # def validate_dir(self):
    #     """Small validation to check if a given path is a valid path in the project.

    #     Raises:
    #         FileNotFoundError: The path does not point to a valid directory.
    #     """
    #     if self.private_wheel_location.is_dir():
    #         log.info(f"Looking for wheel file in {self.private_wheel_location}.")
    #     else:
    #         raise FileNotFoundError

    # def retrieve_whl_filepath(self) -> Path:
    #     """The project inside you develop your model you want to train need to be build as a pip wheel and added to the training environement.

    #     Once your project has been build as a pip wheel (by Flit, setuptools, Hatch, etc), given the location of the distribution directory
    #     (usually `$cwd/dist`), this function looks for a `.whl` file and return its path.

    #     !!! remark "Remark"

    #         Usually, there is only one `.whl` file in a `dist` directory. This is an assumption of this function.

    #     Raises:
    #         FileNotFoundError: Either the `dist` directory path is not valid, or there is no `.whl` file inside.

    #     Returns:
    #         Path: The path of the private wheel package.
    #     """
    #     try:
    #         self.validate_dir()
    #     except FileNotFoundError:
    #         log.error(
    #             f"Couldn't find distribution directory {self.private_wheel_location}",
    #         )
    #         raise

    #     whl_file = sorted(
    #         Path(file)
    #         for file in Path(self.private_wheel_location).glob("**/*.whl")
    #         if file.is_file()
    #     )
    #     if len(whl_file) == 0:
    #         log.error("Couldn't find wheel distribution")
    #         raise FileNotFoundError

    #     log.info(f"Found wheel {self.private_wheel_location / Path(whl_file[0])}")

    #     return self.private_wheel_location / Path(whl_file[0])

    def create_aml_environment(
        self,
        env_name: str,
        description: str,
    ) -> Environment:
        """Create the AzureML Environment once all the requirements have been gathered.

        Args:
            env_name (str): The name of the environment that will be build.
            description (str): Description of the environment.

        Raises:
            ValueError: You have selected something else than "pip", "conda", or "docker" for the `EnvSpecs`.

        Returns:
            Environment: The training environment which will be used.

        !!! remark "Remark"
            The environment returned is just a Dockerfile, it not built as an image and will have to be built during the first training.
            If you want to build it locally, you will have to put `build_locally=True` in `AMLInterface.register_aml_environment`.
        """
        env = Environment(
            build=BuildContext(dockerfile_path=self.dockerfile_path),
            name=env_name,
            description=description,
        )

        return env



from typing import Dict

from azureml.core import Datastore, Workspace
from azureml.core.authentication import ServicePrincipalAuthentication
from azureml.core.compute import AmlCompute, ComputeTarget
from azureml.core.environment import Environment
from azureml.exceptions import ComputeTargetException

from azure_helper.logger import get_logger

log = get_logger()


class AMLInterface:
    def __init__(
        self,
        spn_credentials: Dict[str, str],
        subscription_id: str,
        workspace_name: str,
        resource_group: str,
    ):
        """Instantiate the connection to an Azure Machine Learning Workspace.

        This class is the principal interface with the other ones. It uses a Service Principle **with Password Authentication** connection to interact with Azure
        Machine Learning. The Service Principle credentials needed can be encapsulated in the following Pydantic class.

        ```python
        from pydantic import BaseSettings, Field

        class SpnCredentials(BaseSettings):
            tenant_id: str = Field(env="TENANT_ID")
            service_principal_id: str = Field(env="SPN_ID")
            service_principal_password: str = Field(env="SPN_PASSWORD")
        ```

        !!! info "Information"

            To create a Service Principle with Password Authentication, you can do it using the azure-cli with the following shell command.

            `az ad sp create-for-rbac --name <spn-name>`

            It is important to note down the app id and password from the creation of this service principal! You’ll also need the tenant ID listed here.


        !!! attention "Attention"

            The service principal also needs to have a contributor role in the IAM management of the ressource group.

        It is responsible for :

        * Connect to an existing Storage Account and promote blob inside a specified container into an Azure Machine Learning Datastores.
        * Register Azure Machine Learning Environment created by the [`AMLEnvironment`][azure_helper.steps.create_aml_env] class as Docker Image.
        * Provisioning Compute Instance, either by fetching an existing one or creating a new one.


        Args:
            spn_credentials (Dict[str, str]): Credentials of the Service Principal used to communicate between the different resources of the workspace.
                Must contains the TenantID and the ServicePrincipalID.
            subscription_id (str): The Azure subscription ID containing the workspace.
            workspace_name (str): The workspace name. The name must be between 2 and 32 characters long.
                The first character of the name must be alphanumeric (letter or number), but the rest of the name may
                contain alphanumerics, hyphens, and underscores. Whitespace is not allowed.
            resource_group (str): The resource group containing the workspace.
        """
        auth = ServicePrincipalAuthentication(**spn_credentials)
        self.workspace = Workspace(
            workspace_name=workspace_name,
            auth=auth,
            subscription_id=subscription_id,
            resource_group=resource_group,
        )

    def register_datastore(
        self,
        datastore_name: str,
        container_name: str,
        storage_acct_name: str,
        storage_acct_key: str,
    ):
        """Register an Azure Blob Container as an AZML datastore.

        ie if you have an architecture like the following one.

        ```bash
        Storage_Account : workspaceperso5448820782
            │
            ├── Container : project-mlops-mk-5448820782
            │   ├── blob : train
            │   └── blob : test
        ```

        Then


        ```py
        aml_interface.register_datastore(
            datastore_name="project_name",
            container_name="project-mlops-mk-5448820782",
            storage_acct_name="workspaceperso5448820782",
            storage_acct_key="storage_acct_key",
            )
        ```

        Will register the Blob container `project-mlops-mk-5448820782` as an AZML datastore under the name `project_name`.


        !!! attention "Attention"

            We are talking here about **datastores**, not **datasets**, which are special data assets of a datastore.

            Eg, in the following architecture.

            ```bash
            Storage_Account : workspaceperso5448820782
                │
                ├── Container : project-mlops-mk-5448820782
                │   ├── blob : train
                │   │           ├── x_train.csv
                │   │           └── y_train.csv
                │   └── blob : test
            ```

            `x_train.csv` and `y_train.csv` can be registered as datasets.

        Args:
            datastore_name (str): The name of the AZML datastore, case insensitive, can only contain alphanumeric characters and -.
            container_name (str): The name of the azure blob container.
            storage_acct_name (str): The storage account name.
            storage_acct_key (str): Access keys of your storage account, defaults to None.
        """
        Datastore.register_azure_blob_container(
            workspace=self.workspace,
            datastore_name=datastore_name,
            container_name=container_name,
            account_name=storage_acct_name,
            account_key=storage_acct_key,
        )

    def register_aml_environment(
        self,
        environment: Environment,
        build_locally: bool = False,
    ):
        """Register the environment object in your workspace.

        An environement created by the [`AMLEnvironment`][azure_helper.steps.create_aml_env] class encapsulate in a Docker image
        everything which is needed to make to project work. See its documentation for further explainations.

        Args:
            environment (Environment): A reproducible Python environment for machine learning experiments.
            build_locally (bool, optional): Whether you want to build locally your environment as a Docker image and push the image
                to workspace ACR directly. This is recommended when users are iterating on the dockerfile since local build
                can utilize cached layers. Defaults to False.
        """
        environment.register(workspace=self.workspace)
        if build_locally:
            environment.build_local(
                self.workspace,
                useDocker=True,
                pushImageToWorkspaceAcr=True,
            )

    def get_compute_target(
        self,
        compute_name: str,
        vm_size: str = "",
        min_node: int = 1,
        max_node: int = 2,
    ) -> ComputeTarget:
        """Instantiate a compute instance to train the models.

        If no Compute Instance with the specified parameters is found in the workspace, a new one will be created.

        Args:
            compute_name (str): The name of the compute instance.
            vm_size (str): The size of agent VMs in the the compute instance.
                More details can be found [here](https://aka.ms/azureml-vm-details).
                Note that not all sizes are available in all regions, as detailed in the previous link.
                If not specified, (ie `vm_size = ""`) defaults to `Standard_NC6`.
            min_node (int, optional): The minimum number of nodes to use on the cluster. Defaults to 1.
            max_node (int, optional): The maximum number of nodes to use on the cluster. Defaults to 2.


        Returns:
            ComputeTarget: An instantiated compute instance.
        """
        try:
            compute_target = ComputeTarget(
                workspace=self.workspace,
                name=compute_name,
            )
            log.info("Found existing compute target")
            log.info(
                f"Compute target instantiated : {compute_target.status.serialize()}",
            )
        except ComputeTargetException as err:
            log.warning(
                f"No compute target found. Creating a new compute target. {err}",
            )
            compute_config = AmlCompute.provisioning_configuration(
                vm_size=vm_size,
                min_nodes=min_node,
                max_nodes=max_node,
            )
            compute_target = ComputeTarget.create(
                self.workspace,
                compute_name,
                compute_config,
            )
            compute_target.wait_for_completion(
                show_output=True,
                timeout_in_minutes=10,
            )
            log.info(
                f"Compute target instantiated : {compute_target.status.serialize()}",
            )
        return compute_target



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

import pandas as pd
from azureml.core import Dataset, Datastore, Run
from azureml.core.model import Model
from skl2onnx import __max_supported_opset__, convert_sklearn
from skl2onnx.common.data_types import FloatTensorType
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score

from azure_helper.logger import get_logger

__here__ = os.path.dirname(__file__)

log = get_logger()


class Train(ABC):
    """Abstract class defining what should be the major steps of a training loop.

    Args:
        ABC (class): Abstract Class
    """

    @abstractmethod
    def get_df_from_datastore_path(self, datastore, datastore_path):
        log.info(f"Loading dataset {datastore_path} from datastore {datastore.name}")
        pass

    @abstractmethod
    def prepare_data(self):
        pass

    @abstractmethod
    def train_model(self):
        log.info("Start training model.")
        pass

    @abstractmethod
    def evaluate_model(self):
        log.info("Start evaluating model.")
        pass

    @abstractmethod
    def save_model(self, model):
        log.info("Saving model to ONNX format.")
        pass

    @abstractmethod
    def register_model(self, model_path):
        pass


class TrainingLoopExample(Train):
    def __init__(
        self,
        run: Run,
        trainig_datastore: str,
        model_name: str,
        target_name: str,
        project_name: str,
    ):
        """Typical example of how you could define a training loop to be used in the `ScriptRunConfig` class in the
        [submit_run][azure_helper.steps.create_aml_experiment.AMLExperiment] method.

        you can use it in the following way.

        ```python
        if __name__ == "__main__":

            run = Run.get_context()
            tl = TrainingLoop(
                run=run,
                trainig_datastore="train_datastore",
                model_name="model_name",
                target_name: "y",
                project_name: "project",
            )
            x_train, y_train, x_test, y_test = tl.prepare_data()

            model = tl.train_model(x_train, y_train)
            tl.evaluate_model(model, x_test, y_test)
            model_path = tl.save_model(model)
            tl.register_model(model_path)
        ```

        Args:
            run (Run): The run corresponding to your Experiment.
            trainig_datastore (str): The name of the datastore where you fetch your datasets.
            model_name (str): The name of the model you train.
            target_name (str): The name of the target, in your dataset.
            project_name (str): The name of the project you're working on.
        """

        self.run = run
        self.model_name = model_name
        self.target_name = target_name
        self.project_name = project_name

        self.workspace = run.experiment.workspace
        self.trainig_datastore = trainig_datastore
        self.datastore = Datastore.get(run.experiment.workspace, trainig_datastore)

    def get_df_from_datastore_path(
        self,
        datastore: Datastore,
        datastore_path: str,
    ) -> pd.DataFrame:
        """Utils function to fetch your datas from a datastore.

        Note that this function is different from [`download_blob_to_df`][azure_helper.utils.blob_storage_interface] method.

        `BlobStorageInterface.download_blob_to_df` takes datas from a blob in one of your container located in your
        storage account. We are fetching datas here from a **datastore**, which is a registered blob in yout AZML workspace.

        Obviously, using `get_df_from_datastore_path` on a Datastore or using `BlobStorageInterface.download_blob_to_df`
        on the blob corresponding to the Datastore will get you the same result.

        Args:
            datastore (Datastore): The name of the registered Datastore in your AZML workspace.
            datastore_path (str): The path to the datas you're fetching.

        Returns:
            The fetched datas as a dataframe.
        """
        # In our example we only have single files,
        # but these may be daily data dumps
        log.info(f"Loading dataset {datastore_path} from datastore {datastore.name}")
        datastore_cfg = [(datastore, datastore_path)]
        dataset = Dataset.Tabular.from_delimited_files(
            path=datastore_cfg,
        )
        return dataset.to_pandas_dataframe()

    def prepare_data(self) -> List[pd.DataFrame]:
        """Get all your datas (train, test) at once.

        Returns:
            List[pd.DataFrame]: Your datas.
        """
        x_train = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/train/X_train.csv",
        )

        y_train = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/train/y_train.csv",
        )
        y_train = y_train[self.target_name]

        x_test = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/test/X_test.csv",
        )

        y_test = self.get_df_from_datastore_path(
            self.datastore,
            f"{self.project_name}/test/y_test.csv",
        )
        y_test = y_test[self.target_name]

        return x_train, y_train, x_test, y_test

    def train_model(self, x_train: pd.DataFrame, y_train: pd.DataFrame):
        """Start the training of the model.

        Args:
            x_train (pd.DataFrame): Train dataset.
            y_train (pd.DataFrame): Train target.

        Returns:
            _type_: A trained model.
        """
        log.info("Start training model.")
        model = LogisticRegression()
        model.fit(x_train, y_train)
        return model

    def evaluate_model(self, model, x_test: pd.DataFrame, y_test: pd.DataFrame):
        """Evaluate your model and record the corresponding metric.

        Args:
            model (_type_): The model you want to evaluate.
            x_test (pd.DataFrame): Test/Validation dataset.
            y_test (pd.DataFrame): Test/Validation target.
        """
        log.info("Start evaluating model.")
        y_pred = model.predict(x_test)
        model_f1_score = f1_score(y_test, y_pred)
        self.run.log("F1_Score", model_f1_score)

    def save_model(self, model) -> Path:
        """Convert the model to ONNX and save it.

        Args:
            model (_type_): Your trained model.

        Returns:
            Path: The path where your converted model is located.
        """
        log.info("Saving model to ONNX format.")
        output_dir = Path("outputs")
        output_dir.mkdir(parents=True, exist_ok=True)
        model_path = output_dir / Path("model.onnx")

        initial_types = [("float_input", FloatTensorType([None, model.n_features_in_]))]
        model_onnx = convert_sklearn(
            model,
            initial_types=initial_types,
            target_opset=__max_supported_opset__,
        )

        # Save the model
        with open("outputs/model.onnx", "wb") as f:
            f.write(model_onnx.SerializeToString())
        log.info("Model saved.")
        return model_path

    def register_model(self, model_path: Path):
        """Register your model into your AZML Model Registry.

        Args:
            model_path (Path): The path returned by the function `save_model`.
        """
        self.run.upload_file(str(model_path), "outputs/model.onnx")

        model = self.run.register_model(
            model_name=self.model_name,
            model_path="outputs/model.onnx",
            model_framework=Model.Framework.ONNX,
        )
        self.run.log("Model_ID", model.id)
        log.info(
            f"Model registered with following informations, name : {model.name}, id : {model.id}, version : {model.version}.",
        )



import sys

from loguru import logger

LOG_LVL = "INFO"
APP_LOGGER_NAME = "azure_helper"
LOG_FILE = "./mlops_e2e_pipeline.log"


def get_logger():
    """_summary_

    Returns:
        _type_: _description_
    """

    fmt = (
        "<level>{level: <8}</level> - "
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> - "
        f"{APP_LOGGER_NAME} - "
        "<cyan>{process}</cyan> - "
        "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
        "<level>{message}</level>"
    )
    config = {
        "handlers": [
            {"sink": sys.stdout, "format": fmt, "level": LOG_LVL},
            {"sink": LOG_FILE, "serialize": True, "format": fmt},
        ],
    }
    logger.configure(**config)

    return logger



import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    """
    Make pytest work with loguru. See:
    https://loguru.readthedocs.io/en/stable/resources/migration.html#making-things-work-with-pytest-and-caplog
    """
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)



"""
`init()` and `run()` are both reserved functions with reserved variables needed to deploy a model, whether it is on ACI or AKS.

`init()` defines how to load the model.

`run()` defines how to handle the datas that are fed to the model from the REST endpoint.
"""
import json

import joblib
import numpy as np
from azureml.core.model import Model


def init():
    """_summary_"""
    global model
    model_path = Model.get_model_path("MODEL_NAME")
    model = joblib.load(model_path)


def run(data):
    """_summary_

    Args:
        data (_type_): _description_

    Returns:
        _type_: _description_
    """
    try:
        data = json.loads(data)
        data = data["data"]
        result = model.predict(np.array(data))
        return result.tolist()
    except Exception as e:
        error = str(e)
        return error



import pandas as pd
from pytest import fixture
from sklearn.datasets import make_classification

from azure_helper.steps.create_data import CreateData


@fixture
def train_data():
    x_arr, y_arr = make_classification(
        n_samples=5000,
        n_features=10,
        n_classes=2,
        random_state=1,
    )
    col_names = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
    ]
    x_df = pd.DataFrame(x_arr, columns=col_names)
    y_df = pd.DataFrame({"Target": y_arr})
    # Training set n=3500
    x_train = x_df.iloc[:3500]
    y_train = y_df.iloc[:3500]

    return x_train, y_train


@fixture
def test_data():
    x_arr, y_arr = make_classification(
        n_samples=5000,
        n_features=10,
        n_classes=2,
        random_state=1,
    )
    col_names = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
    ]
    x_df = pd.DataFrame(x_arr, columns=col_names)
    y_df = pd.DataFrame({"Target": y_arr})

    # Testing set n=750
    x_test = x_df.iloc[3500:4250]
    y_test = y_df.iloc[3500:4250]

    return x_test, y_test


@fixture
def val_data():
    x_arr, y_arr = make_classification(
        n_samples=5000,
        n_features=10,
        n_classes=2,
        random_state=1,
    )
    col_names = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
    ]
    x_df = pd.DataFrame(x_arr, columns=col_names)
    y_df = pd.DataFrame({"Target": y_arr})

    # Validation set n=750
    x_val = x_df.iloc[4250:]
    y_val = y_df.iloc[4250:]

    return x_val, y_val


class TestCreateClassificationData:
    def test_init(self, train_data, test_data, val_data):

        x_train, y_train = train_data
        x_test, y_test = test_data
        x_valid, y_valid = val_data

        assert isinstance(x_train, pd.DataFrame)
        assert isinstance(x_test, pd.DataFrame)
        assert isinstance(x_valid, pd.DataFrame)
        assert isinstance(y_train, pd.DataFrame)
        assert isinstance(y_test, pd.DataFrame)
        assert isinstance(y_valid, pd.DataFrame)

        # Train data is 3500 rows
        # X data is 10 columns, y data is 1 column
        assert len(x_train) == 3500
        assert len(x_train.columns) == 10
        assert len(y_train) == 3500
        assert len(y_train.columns) == 1

        # Test data is 750 rows
        assert len(x_test) == 750
        assert len(x_test.columns) == 10
        assert len(y_test) == 750
        assert len(y_test.columns) == 1

        # Validation data is 750 rows
        assert len(x_valid) == 750
        assert len(x_valid.columns) == 10
        assert len(y_valid) == 750
        assert len(y_valid.columns) == 1

    def test_upload_training_data(self, mocker, train_data):

        x_train, y_train = train_data
        mock_blob_storage_interface = mocker.patch(
            "azure_helper.steps.create_data.BlobStorageInterface",
        )

        data_creator = CreateData(
            project_name="test_project",
            train_datastore="train",
            test_datastore="test",
        )

        data_creator.upload_training_data(mock_blob_storage_interface, x_train, y_train)
        assert mock_blob_storage_interface.upload_df_to_blob.call_count == 2

    def test_upload_test_data(self, mocker, test_data):
        x_test, y_test = test_data

        mock_blob_storage_interface = mocker.patch(
            "azure_helper.steps.create_data.BlobStorageInterface",
        )
        data_creator = CreateData(
            project_name="test_project",
            train_datastore="train",
            test_datastore="test",
        )
        data_creator.upload_test_data(mock_blob_storage_interface, x_test, y_test)

        assert mock_blob_storage_interface.upload_df_to_blob.call_count == 2

    def test_upload_validation_data(self, mocker, val_data):
        x_valid, y_valid = val_data

        mock_blob_storage_interface = mocker.patch(
            "azure_helper.steps.create_data.BlobStorageInterface",
        )
        data_creator = CreateData(
            project_name="test_project",
            train_datastore="train",
            test_datastore="test",
        )
        data_creator.upload_validation_data(
            mock_blob_storage_interface,
            x_valid,
            y_valid,
        )

        assert mock_blob_storage_interface.upload_df_to_blob.call_count == 2



from unittest.mock import Mock

from azureml.core import ScriptRunConfig
from azureml.pipeline.core import Pipeline

from azure_helper.steps.create_aml_experiment import AMLExperiment

test_module = "azure_helper.steps.create_aml_experiment"


class TestAMLExperiment:
    def test_submit_run(self, mocker, caplog):
        # We are not testing Azure ML SDK functionality in a unit test
        # Rather that the correct calls are made
        mock_env = mocker.patch(f"{test_module}.Environment")

        # we mock the Experiment class
        mock_experiment = mocker.patch(f"{test_module}.Experiment")
        # this class is instantiate in AMLExperiment and must return an object when submitted
        # so we mock both : the instantiated class and the returned object when submitted
        mock_experiment_obj = Mock()
        mock_run = Mock()
        mock_experiment.return_value = mock_experiment_obj
        mock_experiment_obj.submit.return_value = mock_run

        mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")
        mock_aml_inteface_obj = Mock()
        mock_compute_target = Mock()
        mock_aml_inteface.return_value = mock_aml_inteface_obj
        mock_aml_inteface_obj.get_compute_target.return_value = mock_compute_target

        workspace_name = "test_workspace"
        resource_group = "test_rg"
        subscription_id = "test_sub_id"

        spn_credentials = {
            "tenant_id": "test_tenant_id",
            "service_principal_id": "test_spn_id",
            "service_principal_password": "test_spn_passwd",
        }

        aml_interface = mock_aml_inteface(
            spn_credentials,
            subscription_id,
            workspace_name,
            resource_group,
        )

        aml_exp = AMLExperiment(
            aml_interface=aml_interface,
            aml_compute_name="test_compute_name",
            aml_compute_instance="STANDARD_D2_V2",
            env_name="test_env_name",
            experiment_name="test_exp_name",
            training_script_path="azure_helper/steps/train.py",
            min_node=1,
            max_node=2,
            clean_after_run=True,
        )

        aml_exp.submit_run()

        mock_env.get.assert_called_once()

        mock_experiment_obj.submit.assert_called_once()

        aml_interface.get_compute_target.assert_called_once()

        _, kwargs = mock_experiment_obj.submit.call_args_list[0]
        exp_config = kwargs["config"]
        assert isinstance(exp_config, ScriptRunConfig)
        assert exp_config.script == "azure_helper/steps/train.py"

        mock_run.wait_for_completion.assert_called_once_with(show_output=True)
        mock_run.get_metrics.assert_called_once()

        mock_compute_target.delete.assert_called_once()

        assert len(caplog.records) == 3

    def test_submit_run_no_clean_after_run(self, mocker, caplog):
        # We are not testing Azure ML SDK functionality in a unit test
        # Rather that the correct calls are made

        mock_env = mocker.patch(f"{test_module}.Environment")

        # we mock the Experiment class
        mock_experiment = mocker.patch(f"{test_module}.Experiment")
        # this class is instantiated in AMLExperiment and must return an object when submitted
        # so we mock both : the instantiated class and the returned object when submitted
        mock_experiment_obj = Mock()
        mock_run = Mock()
        mock_experiment.return_value = mock_experiment_obj
        mock_experiment_obj.submit.return_value = mock_run

        mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")
        # this class is instantiated in AMLExperiment and must return an object when submitted
        # so we mock both : the instantiated class and the returned object when submitted
        mock_aml_inteface_obj = Mock()
        mock_compute_target = Mock()
        mock_aml_inteface.return_value = mock_aml_inteface_obj
        mock_aml_inteface_obj.get_compute_target.return_value = mock_compute_target

        workspace_name = "test_workspace"
        resource_group = "test_rg"
        subscription_id = "test_sub_id"

        spn_credentials = {
            "tenant_id": "test_tenant_id",
            "service_principal_id": "test_spn_id",
            "service_principal_password": "test_spn_passwd",
        }

        aml_interface = mock_aml_inteface(
            spn_credentials,
            subscription_id,
            workspace_name,
            resource_group,
        )

        aml_exp = AMLExperiment(
            aml_interface=aml_interface,
            aml_compute_name="test_compute_name",
            aml_compute_instance="STANDARD_D2_V2",
            env_name="test_env_name",
            experiment_name="test_exp_name",
            training_script_path="azure_helper/steps/train.py",
            min_node=1,
            max_node=2,
            clean_after_run=False,
        )

        aml_exp.submit_run()

        mock_env.get.assert_called_once()

        mock_experiment_obj.submit.assert_called_once()

        aml_interface.get_compute_target.assert_called_once_with(
            "test_compute_name",
            "STANDARD_D2_V2",
            1,
            2,
        )

        _, kwargs = mock_experiment_obj.submit.call_args_list[0]
        exp_config = kwargs["config"]
        assert isinstance(exp_config, ScriptRunConfig)
        assert exp_config.script == "azure_helper/steps/train.py"

        mock_run.wait_for_completion.assert_called_once_with(show_output=True)
        mock_run.get_metrics.assert_called_once()

        aml_interface.get_compute_target.assert_called_once()

        mock_compute_target.delete.assert_not_called()

        assert len(caplog.records) == 2

    def test_generate_run_config(self, mocker):
        # We are not testing Azure ML SDK functionality in a unit test
        # Rather that the correct calls are made

        mock_env = mocker.patch(f"{test_module}.Environment")
        mock_env_obj = Mock()
        mock_env.get.return_value = mock_env_obj

        mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")
        # this class is instantiated in AMLExperiment and must return an object when submitted
        # so we mock both : the instantiated class and the returned object when submitted

        workspace_name = "test_workspace"
        resource_group = "test_rg"
        subscription_id = "test_sub_id"

        spn_credentials = {
            "tenant_id": "test_tenant_id",
            "service_principal_id": "test_spn_id",
            "service_principal_password": "test_spn_passwd",
        }

        mock_aml_inteface_obj = mock_aml_inteface(
            spn_credentials,
            subscription_id,
            workspace_name,
            resource_group,
        )

        aml_exp = AMLExperiment(
            aml_interface=mock_aml_inteface_obj,
            aml_compute_name="test_compute_name",
            aml_compute_instance="STANDARD_D2_V2",
            env_name="test_env_name",
            experiment_name="test_exp_name",
            training_script_path="azure_helper/steps/train.py",
            min_node=1,
            max_node=2,
            clean_after_run=False,
        )

        run_config = aml_exp.generate_run_config()

        assert run_config.environment == mock_env_obj
        assert run_config.target == "local"

    def test_submit_ipeline(self, mocker, caplog):
        # We are not testing Azure ML SDK functionality in a unit test
        # Rather that the correct calls are made
        mock_env = mocker.patch(f"{test_module}.Environment")

        mock_pipeline = mocker.patch(f"{test_module}.Pipeline")

        # we mock the Experiment class
        mock_experiment = mocker.patch(f"{test_module}.Experiment")
        # this class is instantiate in AMLExperiment and must return an object when submitted
        # so we mock both : the instantiated class and the returned object when submitted
        mock_experiment_obj = Mock()
        mock_run = Mock()
        mock_experiment.return_value = mock_experiment_obj
        mock_experiment_obj.submit.return_value = mock_run

        mock_aml_inteface = mocker.patch(f"{test_module}.AMLInterface")
        mock_aml_inteface_obj = Mock()
        mock_compute_target = Mock()
        mock_aml_inteface.return_value = mock_aml_inteface_obj
        mock_aml_inteface_obj.get_compute_target.return_value = mock_compute_target

        workspace_name = "test_workspace"
        resource_group = "test_rg"
        subscription_id = "test_sub_id"

        spn_credentials = {
            "tenant_id": "test_tenant_id",
            "service_principal_id": "test_spn_id",
            "service_principal_password": "test_spn_passwd",
        }

        aml_interface = mock_aml_inteface(
            spn_credentials,
            subscription_id,
            workspace_name,
            resource_group,
        )

        aml_exp = AMLExperiment(
            aml_interface=aml_interface,
            aml_compute_name="test_compute_name",
            aml_compute_instance="STANDARD_D2_V2",
            env_name="test_env_name",
            experiment_name="test_exp_name",
            training_script_path="azure_helper/steps/train.py",
            min_node=1,
            max_node=2,
            clean_after_run=True,
        )

        step1 = Mock()
        step2 = Mock()

        aml_exp.submit_pipeline(steps=[step1, step2])

        mock_experiment_obj.submit.assert_called_once()

        aml_interface.get_compute_target.assert_called_once()

        mock_run.wait_for_completion.assert_called_once_with(show_output=True)

        mock_compute_target.delete.assert_called_once()

        assert len(caplog.records) == 3



from unittest.mock import Mock

import pandas as pd
from azure.core.exceptions import ResourceExistsError
from pytest import fixture

from azure_helper.interfaces.blob_storage_interface import BlobStorageInterface

test_module = "azure_helper.interfaces.blob_storage_interface"


@fixture
def blob_storage_resources(mocker):
    blob_service_client_obj = Mock()

    mock_blob_service_client = mocker.patch(f"{test_module}.BlobServiceClient")
    mock_blob_service_client.from_connection_string.return_value = (
        blob_service_client_obj
    )

    blob_storage_interface = BlobStorageInterface(
        "test_storage_acct_name",
        "test_storage_acct_key",
    )

    return mock_blob_service_client, blob_service_client_obj, blob_storage_interface


class TestBlobStorageInterface:
    def test_init(self, blob_storage_resources):

        mock_blob_service_client, _, _ = blob_storage_resources

        mock_blob_service_client.from_connection_string.assert_called_once_with(
            "DefaultEndpointsProtocol=https;"
            + "AccountName=test_storage_acct_name;"
            + "AccountKey=test_storage_acct_key;"
            + "EndpointSuffix=core.windows.net",
        )

    def test_create_container(self, blob_storage_resources, caplog):

        _, blob_service_client_obj, blob_storage_interface = blob_storage_resources

        blob_service_client_obj.create_container.side_effect = [
            None,
            ResourceExistsError,
        ]
        blob_storage_interface.create_container("test_container_name")

        blob_service_client_obj.create_container.assert_called_with(
            "test_container_name",
        )

        # Upon second call to BlobStorageInterface.create_container,
        # ResourceExistsError is raised from
        # BlobStorageInterface.blob_service_client.create_container
        # Second call checks that no error is raised from
        # BlobStorageInterface.create_container
        # with pytest.raises(ResourceExistsError) as err:
        blob_storage_interface.create_container("test_container_name")
        assert len(caplog.records) == 2
        assert (
            "Blob storage container test_container_name already exists"
            in caplog.records[1].message
        )

    def test_upload_df_to_blob(self, blob_storage_resources, caplog):

        _, blob_service_client_obj, blob_storage_interface = blob_storage_resources

        mock_blob_client = Mock()
        mock_blob_client.upload_blob.side_effect = [
            None,
            ResourceExistsError,
            None,
        ]
        blob_service_client_obj.get_blob_client.return_value = mock_blob_client

        test_df = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])

        blob_storage_interface.upload_df_to_blob(
            test_df,
            "test_container_name",
            "test_remote_path",
        )

        assert (
            "Dataset uploaded at blob path : test_remote_path."
            in caplog.records[2].message
        )

        blob_service_client_obj.get_blob_client.assert_called_with(
            container="test_container_name",
            blob="test_remote_path",
        )
        mock_blob_client.upload_blob.assert_called()
        args, _ = mock_blob_client.upload_blob.call_args_list[0]
        upload_arg = args[0]
        # The file uploaded is uploaded as bytes
        assert isinstance(upload_arg, bytes)

        # First time upload_df_to_blob is called, there is no
        # ResourceExistsError so the blob is not deleted
        mock_blob_client.delete_blob.assert_not_called()

        # 2nd call to upload_df_to_blob
        blob_storage_interface.upload_df_to_blob(
            test_df,
            "test_container_name",
            "test_remote_path",
        )
        assert len(caplog.records) == 7
        assert (
            "Blob path test_remote_path already contains datas. Now deleting old datas tu upload the new ones."
            in caplog.records[5].message
        )
        # Second time upload_df_to_blob is called, there is a
        # ResourceExistsError raised so the blob is deleted first
        mock_blob_client.delete_blob.assert_called_once()
        mock_blob_client.upload_blob.assert_called()

    def test_download_blob_to_df(self, blob_storage_resources, caplog):

        _, blob_service_client_obj, blob_storage_interface = blob_storage_resources

        mock_blob_client = Mock()
        mock_stream = Mock()
        test_csv_str = "a,b\n1,2\n3,4"

        blob_service_client_obj.get_blob_client.return_value = mock_blob_client
        mock_blob_client.download_blob.return_value = mock_stream
        mock_stream.content_as_text.return_value = test_csv_str

        output_df = blob_storage_interface.download_blob_to_df(
            "test_container_name",
            "test_remote_path",
        )

        assert isinstance(output_df, pd.DataFrame)
        assert "a" in output_df.columns
        assert output_df.loc[1, "b"] == 4

        blob_service_client_obj.get_blob_client.assert_called_once_with(
            container="test_container_name",
            blob="test_remote_path",
        )
        mock_blob_client.download_blob.assert_called_once()
        mock_stream.content_as_text.assert_called_once()

        assert len(caplog.records) == 2
        assert (
            "Download from test_container_name ended successfully."
            in caplog.records[1].message
        )



from diagrams import Cluster, Diagram, Edge
from diagrams.azure.database import BlobStorage
from diagrams.azure.devops import Pipelines, Repos
from diagrams.azure.general import Resourcegroups
from diagrams.azure.identity import ActiveDirectory
from diagrams.azure.ml import MachineLearningServiceWorkspaces as Workspace
from diagrams.azure.security import KeyVaults
from diagrams.onprem.container import Docker
from diagrams.programming.language import Python

with Diagram(
    "Azure MLOps helper dependency diagram",
    show=False,
    outformat=["svg", "png"],
) as dac:
    dac.dot.renderer = "cairo"

    with Cluster("ML Resources Creation"):

        data = Python("BlobStorageInterface")
        interface = Python("AMLInterface")
        env = Python("AMLEnvironment")

    with Cluster("ML Training"):

        exp = Python("AMLExperiment")

    with Cluster("RG"):
        az = Resourcegroups("Home")
        with Cluster("Workspace"):
            ws = Workspace("Workspace")
            with Cluster("Password Storage"):
                kv = KeyVaults("KeyVault")
                ad = ActiveDirectory("Spn")
                blob = BlobStorage("BlobStorage")

            with Cluster("Assets"):
                ws_env = Workspace("Environments")
                ws_data = Workspace("Datastores")
                ws_exp = Workspace("Experiments")
                ws_model = Workspace("Models")

    with Cluster("Pipeline Store"):
        Repos("Repo")
        with Cluster("Pipelines Templates"):
            Pipelines("data-creation-template")
            Pipelines("env-creation-template")
            Pipelines("experiment-creation-template")

    with Cluster("Repo"):
        Repos("Repo")
        with Cluster("Azure DevOps ML Resources"):
            devops_data = Pipelines("Data Creation")
            devops_env = Pipelines("Environment Creation")
            # devops_exp = Pipelines("Training pipeline")

        with Cluster("Azure DevOps ML Training"):
            # devops_data = Pipelines("Data Creation")
            # devops_env = Pipelines("Environment Creation")
            devops_exp = Pipelines("Training pipeline")

            (
                devops_data
                >> Edge(label="trigger")
                >> devops_env
                >> Edge(label="trigger")
                >> devops_exp
            )
    env << interface
    interface >> Edge(label="register environment") >> ws_env

    data >> Edge(label="data storage") >> blob
    blob << interface
    interface >> Edge(label="register datastore") >> ws_data

    exp >> Edge(label="fetch workspace") >> interface
    exp >> Edge(label="fetch env") >> ws_env
    exp >> Edge(label="submit run") >> ws_exp
    exp >> Edge(label="register model") >> ws_model

    devops_data >> data
    devops_env >> env
    devops_exp >> exp
